#include <format>
#include <Client/LocalWorld.hpp>
#include <JGL/JGL.h>
#include <Core/Data.hpp>
#include <fstream>
#include <Core/Loggers.hpp>
#include <JUI/Widgets/Scene.hpp>

namespace CaveGame::Client {

    LocalWorld::LocalWorld(const std::string& world_name, int seed, bool overwrite)
    : World(world_name, seed, overwrite) {
        Logs::Info("Spinning up ChunkServerThread");
        chunk_thread = std::thread(&LocalWorld::ChunkServerThread, this);
        chunk_thread.detach();
        //hud = nullptr;
    }

    void LocalWorld::PostInit() {
        //font = JGL::Font("assets/fonts/Jupiteroid.ttf");
    }

    void LocalWorld::DrawChunkGrid() const {
        Vector2 viewport_topleft = camera.ScaledViewport().minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewport().maxPoint;

        int nearest_grid_left = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize);
        int nearest_grid_right = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize);

        for (int x = nearest_grid_left; x <= nearest_grid_right; x++) {
            auto top = Vector2(x * Core::Chunk::ChunkSize, viewport_topleft.y);
            auto bottom = Vector2(x * Core::Chunk::ChunkSize, viewport_bottomright.y);
            JGL::J2D::DrawLine(Colors::Red, top, bottom);
        }

        int nearest_grid_top = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize);
        int nearest_grid_bottom = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize);

        for (int y = nearest_grid_top; y <= nearest_grid_bottom; y++) {
            auto left = Vector2(viewport_topleft.x, y * Core::Chunk::ChunkSize);
            auto right = Vector2(viewport_bottomright.x, y * Core::Chunk::ChunkSize);
            JGL::J2D::DrawLine(Colors::Red, left, right);
        }
    }


    void LocalWorld::CheckCachedChunkSprites()
    {
        for (auto&[chunk_pos, chunk] : loaded_chunks) {
            if (IsChunkCellWithinViewport(chunk_pos))
            {
                // No rendertarget for this chunk.
                if (!cached_chunk_sprites.contains(chunk_pos))
                {
                    chunk.touched = false;
                    auto* target = new JGL::RenderTarget({Core::Chunk::ChunkSize, Core::Chunk::ChunkSize}, {0,0,0,0});
                    RenderChunkTexture(chunk_pos, target, chunk);
                    cached_chunk_sprites.insert({chunk_pos, target});
                }
                // rendertarget needs updating.
                else if (chunk.touched) {
                    // TODO: Modify RenderTarget in place.
                    chunk.touched = false;

                    auto* target = cached_chunk_sprites[chunk_pos];

                    RenderChunkTexture(chunk_pos, target, chunk);


                    //cached_chunk_sprites.erase(chunk_pos);
                    //auto* target = new JGL::RenderTarget({Core::Chunk::ChunkSize, Core::Chunk::ChunkSize}, {0,0,0,0});
                    //RenderChunkTexture(chunk_pos, target, chunk);
                    //cached_chunk_sprites.insert({chunk_pos, target});

                }

            }
        }
    }

    void LocalWorld::Draw() {
        CheckCachedChunkSprites();

        JGL::J2D::Begin();


        glClearColor(0.5f,0.5f,1.f,1.f);

        // Shift the origin to the center of the screen.
        glTranslatef(camera.HalfSizeOffset().x, camera.HalfSizeOffset().y, 0);

        // Apply rotation, zoom, and translation.
        glRotatef(camera.Rotation(), 0, 0, 1);
        glScalef(camera.Zoom(), camera.Zoom(), 1);
        glTranslatef(-camera.Position().x, -camera.Position().y, 0);

        // Draw the cached RenderTargets for our chunks.
        for (const auto&[chunk_pos, chunk] : loaded_chunks)
        {
            if (IsChunkCellWithinViewport(chunk_pos))
                RenderChunk(chunk_pos, chunk);

            //JGL::J2D::DrawString(Colors::Black, std::format("{}, {}", chunk_pos.x, chunk_pos.y), chunk.GetChunkRealCoordinates().x, chunk.GetChunkRealCoordinates().y,
            //                     1, 8, font);
        }

        // Debug Grid
        DrawChunkGrid();


        Vector2 transformed = camera.ScreenToWorld(mouse_pos);

        JGL::J2D::OutlineCircle(Colors::Red, transformed, 5);


        // Banana for scale.
        JGL::J2D::FillRect(Colors::Yellow, {0,0}, {32, 48});

        JGL::J2D::End();


        //hud->Draw();
    }

    bool LocalWorld::IsChunkCellWithinViewport(const Vector2 &coords) const {
        AABB2D chunk_bounding_box = AABB2D(coords*Core::Chunk::ChunkSize, coords*Core::Chunk::ChunkSize+Vector2(Core::Chunk::ChunkSize));
        return Core::Solver::AABB2Dvs(
                camera.ScaledViewport(), chunk_bounding_box);
    }

    void LocalWorld::RenderChunkTexture(const Vector2 &coords, JGL::RenderTarget* destination, const Core::Chunk& chunk)
    {
        using CaveGame::Core::TileID;

        std::vector<Vector2> stone_coords;
        std::vector<Vector2> dirt_coords;
        std::vector<Vector2> grass_coords;
        std::vector<Vector2> clay_coords;
        std::vector<Vector2> mud_coords;

        for (int x = 0; x < Core::Chunk::ChunkSize; x++)
        {
            for (int y = 0; y < Core::Chunk::ChunkSize; y++)
            {
                TileID tile = chunk.GetTile(x, y);

                Vector2 relative_tile_coords = Vector2(x, y);
                Vector2 real_tile_coords = relative_tile_coords;

                if (tile == TileID::AIR) // Air
                    continue;


                /*else if (tile == TileID::STONE) // Stone
                    stone_coords.push_back(real_tile_coords);
                else if (tile == TileID::DIRT) // Dirt
                    dirt_coords.push_back(real_tile_coords);
                else if (tile == TileID::GRASS) // Grass
                    grass_coords.push_back(real_tile_coords);
                else if (tile == TileID::CLAY) // Ore
                    clay_coords.push_back(real_tile_coords);
                else if (tile == TileID::MUD) // Water
                    mud_coords.push_back(real_tile_coords);*/
            }
        }
        JGL::J2D::Begin(destination, true);
            JGL::J2D::DrawPoints(Colors::Gray, stone_coords.data(), stone_coords.size());
            JGL::J2D::DrawPoints(Colors::Browns::Chocolate, dirt_coords.data(), dirt_coords.size());
            JGL::J2D::DrawPoints(Colors::Green, grass_coords.data(), grass_coords.size());
            JGL::J2D::DrawPoints(Colors::Reds::Firebrick, clay_coords.data(), clay_coords.size());
            JGL::J2D::DrawPoints(Colors::Browns::BurlyWood, mud_coords.data(), mud_coords.size());
        JGL::J2D::End();
    }

    void LocalWorld::RenderChunk(const Vector2& coords, const Core::Chunk& chunk) {
        auto it = cached_chunk_sprites.find(coords);
        if (it != cached_chunk_sprites.end())
            JGL::J2D::DrawRenderTarget(it->second, chunk.GetChunkRealCoordinates());
    }

    void LocalWorld::LookForChunksNeedUnloading()
    {
        for (auto it = loaded_chunks.begin(); it != loaded_chunks.end();)
        {
            const auto coords = it->first;
            if (!IsChunkCellWithinViewport(coords))
            {

                // TODO: Move off main thread.
                SaveChunkToFile(coords, it->second);

                //delete it->second;
                loaded_chunks.erase(it++);

                cached_chunk_sprites.erase(coords);
            } else
            {
                ++it;
            }
        }
    }

    bool LocalWorld::AwaitingChunkAtCell(const Vector2& cell)
    {
        int cnt = std::count(chunks_in_waiting.begin(), chunks_in_waiting.end(), cell);

        if (cnt > 1)
        {
            // Not good.
            std::cerr << "Fuckiddy fuck" << std::endl;
            return false;
        }
        if (cnt == 1)
            return true;
        return false;
    }

    void LocalWorld::LookForChunksNeedLoading()
    {

        // TODO: Implement mechanism to load chunk on separate thread.
        // TODO: ConcurrentQueue<Vector2> RequestedChunks;
        // TODO: Chunk Generator / Loader From File on separate thread.

        // TODO: Make sure an extra layer of chunks beyond the viewport is generated.
        // This will limit the amount of "pop-in" the player will notice.

        Vector2 viewport_topleft = camera.ScaledViewport().minPoint;
        Vector2 viewport_bottomright = camera.ScaledViewport().maxPoint;

        int lower_bound_h = Math::Floor(viewport_topleft.x / Core::Chunk::ChunkSize);
        int upper_bound_h = Math::Floor(viewport_bottomright.x / Core::Chunk::ChunkSize);

        int lower_bound_v = Math::Floor(viewport_topleft.y / Core::Chunk::ChunkSize);
        int upper_bound_v = Math::Floor(viewport_bottomright.y / Core::Chunk::ChunkSize);

        for (int x = lower_bound_h; x <= upper_bound_h; x++)
        {
            for (int y = lower_bound_v; y <= upper_bound_v; y++)
            {
                Vector2 cell = Vector2(x,y);
                if (!HasChunkAtCell(cell) || !AwaitingChunkAtCell(cell))
                    RequestChunk(cell);

            }
        }
    }

    void LocalWorld::Update(float elapsed) {
        World::Update(elapsed);

        camera.Update(elapsed);

        LookForChunksNeedUnloading();
        LookForChunksNeedLoading();



        // TODO: Unload chunks once they're offscreen for a certain time threshold.


        // Check our generator queue for complete chunks, and pull them into our loaded chunks.
        while (!ServedChunks.empty())
        {
            Core::Chunk c = Core::Chunk({0,0});
            ServedChunks.front_pop(c);
            loaded_chunks.emplace(c.GetChunkCell(), c);

            if (std::find(chunks_in_waiting.begin(), chunks_in_waiting.end(), c.GetChunkCell()) != chunks_in_waiting.end()) {
                chunks_in_waiting.erase(std::remove(chunks_in_waiting.begin(), chunks_in_waiting.end(), c.GetChunkCell()), chunks_in_waiting.end());
            }
        }
    }

    void LocalWorld::RequestChunk(const Vector2 &cell) {
        // TODO: Ensure we cannot request the same chunk twice..
        if (std::find(chunks_in_waiting.begin(), chunks_in_waiting.end(), cell) == chunks_in_waiting.end())
        {

             RequestedChunks.push(cell);
             chunks_in_waiting.push_back(cell);
        }
    }

    void LocalWorld::ChunkServerThread() {
        while (run_chunk_thread)
        {
            while (!RequestedChunks.empty())
            {
                Vector2 cell;
                RequestedChunks.front_pop(cell);

                if (HasChunkOnFile(cell)) {
                    ServedChunks.emplace(Core::Chunk(cell, GetChunkFullPath(cell)));
                } else {
                    Core::Chunk chunk(cell);
                    generator.FirstPass(chunk);
                    //SaveChunkToFile(cell, chunk);
                    ServedChunks.emplace(chunk);
               }


                //generator.FirstPass()
                //std::this_thread::sleep_for(100ms);
            }
            //std::this_thread::sleep_for(250ns);
        }
    }


    void LocalWorld::RefreshAll() {
        //World::RefreshAll();
        cached_chunk_sprites.clear();
        chunks_in_waiting.clear();
    }


}

