/// CaveGame - A procedural 2D platformer sandbox.
/// Created by Josh O'Leary @ Redacted Software, 2020-2024
/// Contact: josh@redacted.cc
/// Contributors: william@redacted.cc maxi@redacted.cc
/// This work is dedicated to the public domain.

/// @file Chunk.hpp
/// @desc A subdivision of the gameworld into discretized blocks.
/// @edit 12/2/2024

#pragma once

#include <J3ML/Geometry/QuadTree.hpp>
#include "Tile.hpp"
#include "Data.hpp"
#include <filesystem>
#include <fstream>

namespace CaveGame::Core
{

    class Chunk {
    public:
        static constexpr int ChunkSize = 256;
    public:
        /// The default constructor does not initialize any members.
        Chunk() {}


        ~Chunk()
        {

        }

        /// Constructs a chunk with empty (air) tiles, located at the given chunk-cell.
        /// @param cell_coords Specifies the chunk-cell this chunk will occupy.
        explicit Chunk(const Vector2 &cell_coords);

        explicit Chunk(const Vector2 &cell_coords, const std::filesystem::path &file) : Chunk(cell_coords)
        {
            std::ifstream inp;

            inp.open(file, std::ios::binary | std::ios::in);
            inp.seekg(0, std::ios::end);
            int length = inp.tellg();
            inp.seekg(0, std::ios::beg);

            char buffer[length];
            inp.read(buffer, length);
            inp.close();

            SetData(buffer);
        }

        TileID GetTile(int x, int y) const;
        void SetTile(int x, int y, TileID t);

        [[nodiscard]] Vector2 GetChunkCell() const;
        [[nodiscard]] Vector2 GetChunkRealCoordinates() const;

        [[nodiscard]] const TileID* ptr() const { return &tiles[0][0];}


        [[nodiscard]] std::vector<TileID> DataContiguous() const
        {
            std::vector<TileID> data;
            data.reserve(ChunkSize*ChunkSize);
            for (int x = 0; x < ChunkSize; x++) {
                for (int y = 0; y < ChunkSize; y++) {
                    data.push_back(tiles[x][y]);
                }
            }
            return data;
        }

        void SetData(char* buffer)
        {
            auto* data = reinterpret_cast<TileID*>(buffer);


            for (int x = 0; x < ChunkSize; x++)
            {
                for (int y = 0; y < ChunkSize; y++)
                {
                    tiles[x][y] = data[y+(x*ChunkSize)];
                }
            }
        }

        [[nodiscard]] static constexpr std::size_t BufferSizeBytes()
        {
            return (ChunkSize*ChunkSize)*sizeof(TileID);
        }

        [[nodiscard]] static constexpr std::size_t BufferSize()
        {
            return (ChunkSize*ChunkSize);
        }
    public:
        bool touched;
    protected:
        Vector2 cell;
        TileID tiles[ChunkSize][ChunkSize];

    };
}