#pragma once

#include <cstdint>
#include <Color4.hpp>
#include <Colors.hpp>
#include <Core/Data.hpp>
#include <Core/Registry.hpp>
#include <functional>
#include <Core/Interfaces.hpp>
#include <J3ML/LinearAlgebra/Vector2i.hpp>

namespace CaveGame::Core
{
    using J3ML::LinearAlgebra::Vector2i;

    /*
     *
     *  Tile and TileTrait example structure
     *  - Maxine
     *
     */

    /*
    class TileTrait
    {
    public:
        virtual std::string MnemonicID() = 0;
        virtual void Invoke() = 0;
    };

    class Tile
    {
    public:
        virtual std::string MnemonicID() = 0;
        virtual std::vector<TileTrait> Traits() = 0;
        virtual void AddTrait(TileTrait t) = 0;
        virtual void DelTrait(TileTrait t) = 0;
    };

    class Spreadable : public TileTrait {
    public:
        std::string MnemonicID() { return "tr_spreadable"; };
        void Invoke() {
            // Spread to adjacent tiles using calculation
            calc(world, owner);
        };

    public:
        using Calculator = std::function<void(ITileMap *w, Tile* o)>;
        Spreadable(ITileMap* w, Tile* o, Calculator c) : TileTrait() {
            world = w;
            owner = o;
            calc = c;
        };

    private:
        ITileMap* world;
        Tile* owner;
        Calculator calc;
    };

    class DecayTo : public TileTrait {
    public:
        std::string MnemonicID() { return "tr_decayto"; };
        void Invoke() {
            // Decay to tile on condition
            if (cond(owner))
                perf(world, owner);
        };

    public:
        using Condition = std::function<bool(Tile* o)>;
        using Perform = std::function<void(ITileMap* w, Tile* o)>;
        DecayTo(ITileMap* w, Tile* o, Condition c, Perform p) : TileTrait() {
            world = w;
            owner = o;
            cond = c;
            perf = p;
        }

    private:
        ITileMap* world;
        Tile* owner;
        Condition cond;
        Perform perf;
    };
     */

    enum TileFlags : uint16_t
    {
        SOLID,
        NONSOLID = !SOLID,

    };

    // Use this for space-efficiency and convert to Color3 for doing color manipulations.
    // Struct allocation is cheaper.
    struct Light3
    {
        uint8_t r;
        uint8_t b;
        uint8_t g;
    };


    class TileGroup {};

    class Tile
    {
    public:
        TileID numeric_id = TileID::VOID;
        std::string mnemonic_id;
        Color4 base_color;
        std::vector<Color4> color_pallet;
        bool collides;
        bool does_random_ticc;
    public:

        Tile(TileID numeric, const std::string& name, Color4 color)
        : numeric_id(numeric), mnemonic_id(name), base_color(color)
        {}
        virtual TileID NumericID() const {  return numeric_id; };
        virtual std::string MnemonicID() const {  return mnemonic_id; };
        bool DoesRandomTicc() const { return does_random_ticc; }


        virtual Color4 ComputeColor(TileState state, int x, int y) {}
        virtual void ForcedTicc(ITileMap* world, TileState state, int x, int y) {}
        virtual void RandomTicc(ITileMap* world, TileState state, int x, int y) {}


        template <TileID TDecaysTo>
        void DecayTo(ITileMap *world, TileState state, int x, int y)
        {
            world->SetTile(x, y, TDecaysTo);
        }

        void DecayTo(ITileMap *world, TileState state, int x, int y, TileID TDecaysTo)
        {
            world->SetTile(x, y, TDecaysTo);
        }

    };


    class SoilTile : public Tile
    {
    public:
        TileID decays_to;
        SoilTile(TileID numeric, const std::string& name, Color4 color, TileID decays_target)
        : Tile(numeric, name, color)
        {
            decays_to = decays_target;
            does_random_ticc = true;
        }


        void RandomTicc(ITileMap *world, TileState state, int x, int y) override
        {
            if (!world->HasAdjacentOrDiagonalAirBlock(x, y))
                DecayTo(world, state, x, y, decays_to);
        }

        //void CanGrowGrass
    };

    class DirtTile : public SoilTile
    {

    };

    class GrassyTile : public SoilTile
    {
    public:
        GrassyTile(TileID numeric, const std::string& name, Color4 color)
        : SoilTile(numeric, name, color, TileID::DIRT)
        {
            does_random_ticc = true;
        }

        void RandomTicc(ITileMap *world, TileState state, int x, int y) override
        {
            if (!world->HasAdjacentOrDiagonalAirBlock(x, y))
            {
                DecayTo(world, state, x, y, decays_to);
                return;
            }

            world->GrassRandomTicc(x, y);
        }
    };

    class GasTile : public Tile
    {

    };


    // TODO: Constexpr-qualify Color4



    void RegisterTile(Tile* data);

    Tile* GetByNumeric(TileID id);

    Tile* GetByName(const std::string& name);

    void DefineTiles();

}