#include <map>
#include "Chunk.hpp"
#include "Generator.hpp"
#include "J3ML/LinearAlgebra/Vector2i.hpp"
#include "Interfaces.hpp"
#include <unordered_map>
#include <filesystem>
#include <future>


#pragma once


/// Implements a naiive-AF hashing algorithm for Vector2 class
/// So we can use it as unique keys in std::unordered_map
template<>
struct std::hash<Vector2>
{
    std::size_t operator()(const Vector2& k) const
    {
        return Math::Pow(k.x, 3) + Math::Pow(k.y, 2);
    }
};

namespace CaveGame::Core
{

    class WorldAsyncIOThreadManager
    {

    };

    class World : public ITileMap {
    public:
        constexpr static uint RandomTileTickCoefficient = 100;
        constexpr static float PeriodicAutosaveIntervalSeconds = 30.f;
        World() = default;
        ~World() = default;

        explicit World(const std::string &worldName, int seed = 0, bool overwrite = false);

        TileID GetTile(int x, int y) const override;

        void SetTile(int x, int y, TileID t) override;

        TileState GetTileState(int x, int y) const override { return 0; /* TODO: Implement tile state field */ }

        void SetTileState(int x, int y, TileState state) { /* TODO: Implement tile state field */ }

        // TODO: Doesn't really belong here.
        Vector2 ToUnitDirection(float rotation);

        Chunk GetChunkAtCell(const Vector2 &cell);

        std::unordered_map<Vector2, Chunk> GetChunkList();

        void DoRandomTileTicks();

        virtual void Update(float elapsed);

        virtual void RefreshAll();

        virtual void AutoSaveAsync() { }

        virtual void AutoSave();

        virtual void SaveAndExit();

        /// Returns whether or not the world currently has a chunk loaded at the specified chunk coordinates.
        /// @note Valid Chunk coordinates are integers only.
        bool HasChunkAtCell(const Vector2 &cell) const;

        bool HasChunkOnFile(const Vector2 &cell) const;

        std::filesystem::path GetWorldRootPath() const;

        std::filesystem::path GetChunkFullPath(const Vector2 &cell) const;

    protected:
        void SaveChunkToFile(const Vector2 &cell, const Chunk &chunk);
        std::future<bool> SaveChunkToFileAsync(const Vector2& cell, const Chunk& chunk);
        bool ValidCoords(const Vector2 &cell) const;
    protected:
        const std::filesystem::path worlds {"worlds"};
        std::unordered_map<Vector2, Chunk> loaded_chunks;
        Generator generator;
        float time_of_day;
        std::string world_name;
        RNG rng;
        float autosave_timer = 0.f;

    private:



    };
}