#include <Core/Generator.hpp>
#include <Core/Data.hpp>

namespace CaveGame::Core
{
    Generator::Generator(int seed) : perlin(seed), rng(seed)
    {

        this->seed = seed;
        simplex = SimplexNoise();

        for (int x = 0; x < PrecomputedWhiteNoiseResolution; x++)
        {
            precomputed_white_noise_1D[x] = rng.Float01Incl();

            for (int y = 0; y < PrecomputedWhiteNoiseResolution; y++)
            {
                precomputed_white_noise_2D[x][y] = rng.Float01Incl();
            }
        }
    }

    float Generator::ComputeDepth(int wx, int wy)
    {
        int tile_id = 0;

        // this looks almost the same as the octave noise but with significantly less expense?
        //float height_map_low_pass = perlin.Noise(wx / HeightMapHighPassScale, 0.f, 0.25f) * HeightMapHighPassAmplitude;

        // TODO Use threading to run the high pass and low pass at the same time because it's expensive.
        float height_map_low_pass = perlin.Noise(3, wx / HeightMapHighPassScale, 0.f)* HeightMapHighPassAmplitude;
        float height_map_high_pass = perlin.Noise(wx / HeightMapLowPassScale, 0.f, 0.5f)* HeightMapLowPassAmplitude;


        float height_map = height_map_low_pass + height_map_high_pass;
        return wy - (BaseSurfaceLvl+height_map);
    }

    TileID Generator::HeightMap(float depth, int wx, int wy) {

        float rand = GetPrecomputedWhiteNoise2D(wx, wy)*(TopSoilDepth/2.f);

        if (wy > 1000)
            return TileID::STONE;

        if (depth < 0)  { return TileID::AIR; }
        else if (depth < 2.5f) { return TileID::GRASS; }
        else if (depth < TopSoilDepth-rand) { return TileID::DIRT; }
        else { return TileID::STONE; }
    }



    TileID Generator::ComputeTile(int wx, int wy) {
        float depth = ComputeDepth(wx, wy);
        TileID base = HeightMap(depth, wx, wy);

        /*float cave_high_sx = 48.f;
        float cave_high_sy = 48.f;
        float cave_low_sx = 192.f;
        float cave_low_sy = 192.f;
        float cave_high_pass = perlin.Noise(wx / cave_high_sx, wy / cave_high_sy, 1.f);
        float cave_low_pass = perlin.Noise(wx / cave_low_sx, wy / cave_low_sy, 0.5f);

        float cave_sum = cave_high_pass + cave_low_pass;
        float cave_product = cave_high_pass * cave_low_pass;

        if (cave_sum > -0.05f && 0.2f > cave_sum && cave_product > -0.15f && 0.15f > cave_product)
        {
            return Core::AIR;
        } else { return base; }*/


        float cave_erosion = perlin.Noise(wx / CaveInputScale, wy / CaveInputScale, 0.f) * CaveOutputScale;

        float cave_addition = perlin.Noise(wx / CaveAdditiveInputScale, wy / CaveAdditiveInputScale, 0.f)*CaveAdditiveOutputScale;
        cave_erosion += cave_addition;

        // Make cave openings more narrow
        if (depth < 100) {
            cave_erosion *= 2.f;
        }

        if (cave_erosion > -CaveErosionRange && CaveErosionRange > cave_erosion)
        {
            return TileID::AIR;
        } else { return base; }

    }

    void Generator::FirstPass(CaveGame::Core::Chunk &chunk) {
        Vector2 real_coords = chunk.GetChunkRealCoordinates();

        for (int x = 0; x < Chunk::ChunkSize; x++) {
            for (int y = 0; y < Chunk::ChunkSize; y++) {

                int wx = real_coords.x + x;
                int wy = real_coords.y + y;

                chunk.SetTile(x, y, ComputeTile(wx, wy));
            }
        }
    }

    float Generator::GetPrecomputedWhiteNoise2D(int x, int y) const {
        x = ModInt(x, PrecomputedWhiteNoiseResolution);
        y = ModInt(y, PrecomputedWhiteNoiseResolution);
        return precomputed_white_noise_2D[x][y];
    }

    float Generator::GetPrecomputedWhiteNoise1D(int x) const {
        x = ModInt(x, PrecomputedWhiteNoiseResolution);
        return precomputed_white_noise_1D[x];
    }

}
