#include <Core/World.hpp>
#include <filesystem>
#include <iostream>
#include <fstream>
#include <format>
#include <chrono>
#include <Core/Loggers.hpp>

#include <JJX/JJX.hpp>

namespace CaveGame::Core
{
    void CaveGame::Core::World::Update(float elapsed) {

        autosave_timer += elapsed;

        if (autosave_timer >= PeriodicAutosaveIntervalSeconds)
        {
            AutoSave();
            autosave_timer = 0;
        }

        DoRandomTileTicks();
    }

    TileID World::GetTile(int x, int y) const {
        float chunkX = Math::Floor((float)x / Chunk::ChunkSize);
        float chunkY = Math::Floor((float)y / Chunk::ChunkSize);

        float tileX = Math::Mod(x, Chunk::ChunkSize);
        float tileY = Math::Mod(y, Chunk::ChunkSize);

        if (tileX < 0)
            tileX = Chunk::ChunkSize + tileX;
        if (tileY < 0)
            tileY = Chunk::ChunkSize + tileY;

        Vector2 coords = Vector2(chunkX, chunkY);


        if (HasChunkAtCell(coords))
            return loaded_chunks.at(coords).GetTile(tileX, tileY);

        return TileID::VOID;
    }

    void World::SetTile(int x, int y, TileID t) {
        float chunkX = Math::Floor((float)x / Chunk::ChunkSize);
        float chunkY = Math::Floor((float)y / Chunk::ChunkSize);

        float tileX = Math::Mod(x, Chunk::ChunkSize);
        float tileY = Math::Mod(y, Chunk::ChunkSize);

        if (tileX < 0)
            tileX = Chunk::ChunkSize + tileX;
        if (tileY < 0)
            tileY = Chunk::ChunkSize + tileY;


        const Vector2 coords = Vector2(chunkX, chunkY);

        if (HasChunkAtCell(coords))
            loaded_chunks.at(coords).SetTile(tileX, tileY, t);

    }

    Chunk World::GetChunkAtCell(const Vector2 &cell) {
        if (ValidCoords(cell))
            return loaded_chunks.at(cell);
    }

    bool World::ValidCoords(const Vector2& cell) const {
        if (Math::Floor(cell.x) != cell.x || Math::Floor(cell.y) != cell.y) {
            std::cerr << "Invalid cell coordinate: " << cell << ", must be whole numbers!";
            return false;
        }

        return true;
    }

    bool World::HasChunkAtCell(const Vector2 &cell) const {
        if (ValidCoords(cell))
            return loaded_chunks.contains(cell);
        return false;
    }

    std::unordered_map<Vector2, Chunk> World::GetChunkList() { return loaded_chunks; }

    Vector2 World::ToUnitDirection(float rotation) {
        return {std::cos(rotation), std::cos(rotation)};
    }


    std::string read_text_file(const std::filesystem::path& path)
    {
        std::ifstream file(path, std::ios::binary);
        if (!file)
            throw std::runtime_error("We couldn't find the file: " + path.string());

        std::streamsize file_size;
        file.seekg(0, std::ios::end);
        file_size = file.tellg();
        file.seekg(0, std::ios::beg);

        std::string file_content(file_size, '\0');
        file.read(&file_content[0], file_size);
        file.close();

        return file_content;

    }

    void write_text_file(const std::filesystem::path& path, const std::string& text)
    {
        std::ofstream file(path, std::ios::binary);
        if (!file)
            throw std::runtime_error("We couldn't find the file: " + path.string());

        file << text;
        file.close();
    }


    std::string timestamp()
    {
        return std::format("{:%F %T}", std::chrono::system_clock::now());
    }

    World::World(const std::string &worldName, int seed, bool overwrite) : world_name(worldName), generator(seed) {

        namespace fs = std::filesystem;
        namespace json = JJX::json;

        /// Create "worlds" directory if not already there.
        if (!fs::exists(worlds))
            fs::create_directory(worlds);

        fs::path current_world_path = worlds/worldName;

        if (overwrite)
        {
            if (fs::exists(current_world_path))
                fs::remove_all(current_world_path);
        }

        /// Create directory for this world if not already there.
        if (!fs::exists(current_world_path))
            fs::create_directory(current_world_path);

        /// Create directory for chunks if not already there.
        if (!fs::exists(current_world_path/"chunks"))
            fs::create_directory(current_world_path/"chunks");

        fs::path metadata_path = current_world_path/"meta.json";

        if (!fs::exists(metadata_path)) {
            /// Fill metadata file with base if it is not present.

            auto world_meta_template = json::object({
                {"time_of_day", json::number(time_of_day)},
                {"seed", json::number(seed)},
                {"name", json::string(worldName)},
                {"creation_timestamp", json::string(timestamp())}
            });

            write_text_file(metadata_path, json::deparse(world_meta_template));
        } else
        {
            auto[val, errcode] = json::parse(read_text_file(metadata_path));
        }


    }

    bool World::HasChunkOnFile(const Vector2 &cell) const {
        return std::filesystem::exists(worlds/world_name/"chunks"/cell.ToString());
    }

    std::filesystem::path World::GetWorldRootPath() const
    {
        return worlds/world_name;
    }

    std::filesystem::path World::GetChunkFullPath(const Vector2& cell) const
    {
        return GetWorldRootPath()/"chunks"/cell.ToString();
    }

    void World::SaveChunkToFile(const Vector2& cell, const Chunk& chunk)
    {
        std::ofstream fout;
        fout.open(GetChunkFullPath(cell), std::ios::out | std::ios::binary);
        fout.write(reinterpret_cast<const char*>(chunk.ptr()), chunk.BufferSizeBytes());

        fout.close();
    }

    void World::RefreshAll() {
        loaded_chunks.clear();
    }



    void World::DoRandomTileTicks() {
        // Every frame, In each loaded chunk, perform a RandomTileTick on N tiles.
        // Where N is RandomTileTickCoefficient.

        int max_tries = RandomTileTickCoefficient*5;
        int max_successes = RandomTileTickCoefficient;

        Tile* tile = nullptr;

        for (auto& [coords, chunk] : loaded_chunks)
        {
            for (int ticc = 0; ticc < RandomTileTickCoefficient; ticc++)
            {
                // Select random x,y coordinates between [0, ChunkSize]
                int x = rng.Int(0, Chunk::ChunkSize-1);
                int y = rng.Int(0, Chunk::ChunkSize-1);

                int wx = coords.x*Chunk::ChunkSize + x;
                int wy = coords.y*Chunk::ChunkSize + y;

                TileID at = chunk.GetTile(x, y);

                if (at == TileID::AIR)
                    continue;

                bool slow_method = true;

                // TODO: Unacceptably slow!!!
                if (slow_method)
                {
                    tile = GetByNumeric(at);

                    if (tile != nullptr)
                        if (tile->DoesRandomTicc())
                            tile->RandomTicc(this, 0, wx, wy);
                } else {
                    // Fast method:
                    if (at == TileID::GRASS)
                        GrassRandomTicc(wx, wy);
                }
            }
        }
    }

    void World::AutoSave() {
        Logs::Info("Autosaving...");
        for (auto&[coords, chunk] : loaded_chunks)
        {
            SaveChunkToFile(coords, chunk);
        }
        Logs::Info("Autosave successful!");
    }

    void World::SaveAndExit() {
        Logs::Info("Saving before exiting...");
        for (auto&[coords, chunk] : loaded_chunks)
        {
            SaveChunkToFile(coords, chunk);
        }
        Logs::Info("Saving successful!");
    }


}
